#include	<iostream>
#include	<string.h>
#include	<stdlib.h>
#include	"fsa.h"
#include	"nnode.h"

#define fsa_set_last(arc, f) if (f) arc[goto_offset] |= 2; else arc[goto_offset] &= 0xfd
#define fsa_set_final(arc, f) if (f) arc[goto_offset] |= 1; else arc[goto_offset] &= 0xfe
#define fsa_set_next(arc, f) if (f) arc[goto_offset] |= 4; else arc[goto_offset] &= 0xfb

const int	INDEX_SIZE_STEP = 16;	// allocate chunks of 16 pointers

int	node::no_of_arcs = 1;	// initially there is an empty arc #0

using namespace std;

int	node::next_nodes = 0;

/* Purpose:	Checks whether the last arc of this node leads to another node.
 * Returns:	TRUE if the arcs leads to a node, FALSE otherwise.
 * Remarks:	Node with no arcs is infertile. */
int node::fertile(void) const { return no_of_children && children[no_of_children - 1].child ? TRUE : FALSE; }

/* Purpose:	Find if there is a subgraph in the automaton that is isomorphic
 *		to the subgraph that begins at the last child of this node.
 * Returns:	the node.
 * Remarks:	Because the node can be deleted, and replaced by a pointer
 *		to another node, the method must be invoked with the parent
 *		of the node in question, so that the pointer can be modified.
 *
 *		hit_count counts the number of (registered) links that lead
 *		to a node. hit_count = 0 means the node is not registered yet.
 *		hit_count is also used to avoid deleting nodes that are
 *		pointed to somewhere else in the automaton. */
node * node::compress_or_register(void) {
  node		*c_node;		// current node
  node		*new_node;

  if (no_of_children) {
    c_node = children[no_of_children - 1].child;
    // Nothing to be registered. One letter has been previously added
    // to this node, which resulted only in creation of another arc,
    // but there is no new node attached to it.
    if (c_node == NULL) return this;
    if (c_node->hit_count == 0) {
      // node has not been registered yet
      if (c_node->no_of_children) {
	// compress or register the children (i.e. the line of the youngest)
	// NOTE: the node may have children, but this only means that there
	//       are arcs going from that node; there is no guarrantee
	//       there are nodes at the end of those arcs.
	c_node->compress_or_register();

	// Check if there is already an isomorphic subgraph in the automaton
	if ((new_node = find_or_register(c_node, PRIM_INDEX, REGISTER))) { // Isomorphic subgraph found
	  delete_branch(c_node); // Delete c_node
	  children[no_of_children - 1].child = new_node; // Link found node to parent
	}
	else new_node = c_node; // subgraph is unique, registered
	new_node->hit_count++;	// count number of links to that node
      }
    }//if node not registered previously
  }//if there is a node to register

  return this;
}

/* Purpose:	Computes a hash function for the node to be used in index.
 * Parameters:	start		- (i) first arc number
 *		how_many	- (i) how many arcs are to be considered.
 * Returns:	Hash function.
 * Remarks:	Value of hash function depends on labels of arcs going from this node. */
int node::hash(const int start, const int how_many) const {
  long	h = 0;
  int	finish = start + how_many;
  for (int i = start; i < finish; i++) h = (((h * 17) ^ children[i].letter) ^ (long(children[i].child) >> 2));
  return (h & MAX_ARCS_PER_NODE);
}

/* Purpose:	Assign an index to each arc in the automaton.
 * Parameters:	gtl		- (i) length of the goto field or 0 meaning count arcs not bytes.
 * Returns:	Number of the first arc of the node.
 * Remarks:	Arc numbers are kept in nodes, rather than in arcs themselves.
 *		A node keeps an arc number of its first arc.
 *		Arc number equal -1 means the node has not been visited yet.
 *		Arcs of the root are numbered from 1, because arc #0 is
 *		a dummy arc.
 *
 *		The pictures below describe the situation without STOPBIT:
 *
 *		In pictures below, "->-" is a link to big_brother, either
 *		a true big_brother (i.e. one that is bigger), or the first
 *		node of merged nodes. "-=-" indicates a link to a node
 *		that this node was merged with. Index n indicates already
 *		numbered node. Xo means brother_offset of X. a means
 *		current arc number to be allocated.
 *
 *		Situations and actions (this (A) is always not numbered):
 *
 *		1) A			- number A and children of A
 *					  Note: A may be a big brother
 *					  of another node. It does not matter.
 *					  A = a.
 *
 *		2a) A ->-> B		- number B, number A, and number
 *					  children of B.
 *					  B = a; A = B + Ao.
 *
 *		2b) A ->-> Bn		- number A.
 *					  A = B + Ao.
 *
 *		3a) A ->-> B		- number B, number A, number children
 *		      <-=-		  of B, number children[1] of A.
 *					  B = a; A = B + Ao;
 *
 *		3b) A -=-> B		- number A, number B, number children
 *		      <-<-		  of A, number children[1] of B
 *					  Note: in 3, A & B are always numbered
 *					  at the same time.
 *					  A = Ao; B = A + Bo.
 *
 *		4a) A ->-> B ->-> C	- number C, number B, number A,
 *			     <-=-	  number children of C,
 *					  number children[1] of B.
 *					  C = a; B = C + Bo; A = B + Ao.
 *
 *		4b) A ->-> B -=-> C	- number B, number C, number A,
 *			     <-<-	  number children of B,
 *					  number children[1] of C.
 *					  B = a; C = B + Co; A = B + Ao;
 *
 *		4c) A ->-> Bn ->-> Cn	- number A.
 *			      <-=-	  A = B + Ao.
 *
 *		4d) A ->-> Bn -=-> Cn	- number A.
 *			      <-<-	  A = B + Ao.
 *
 *		With STOPBIT, the situation is quite less complicated,
 *		as JOIN_PAIRS cannot be on. */
int node::number_arcs(const int gtl) {
  node		*p, *pp;
  int		limit;
  // Set the offset for the goto field, i.e. how far the pointer is
  // from the beginning of the arc/transition
//  const int	goto_offset = 1;

  // Number arcs - or actually, the first arc in a node
  if (no_of_children) {
    p = this;
    limit = no_of_children;

    if (big_brother) {
      if (big_brother->arc_no != -1) {
	  // big brother has been numbered
	  // Situations 2b, 4c, 4d.
	  arc_no = (goto_offset + gtl) * brother_offset + big_brother->arc_no;
	  return arc_no;
	  // Note: no need to number children (already numbered)
      }
      else {
	// Big brother not numbered yet

	  // Situation 2a
	  p = big_brother;
	  arc_no = (gtl ? (goto_offset + gtl) : 1) * brother_offset + no_of_arcs;
	  limit = p->no_of_children;
	  // handled: An.
	  // Wait for: Bn, Bc.
      }//if the big brother is not numbered yet
    }//if node has big brother

    p->arc_no = no_of_arcs;
    // All Xn handled.

    // update the arc counter, so that new arcs receive proper numbers
    if (gtl) {
      // We recalculate the numbers taking variable arc length into account
      no_of_arcs += limit * (goto_offset + gtl);
      if (gtl > 1 && no_of_children && nis_next_node(p->children[p->no_of_children - 1].child)) no_of_arcs -= (gtl - 1);
    }
    else no_of_arcs += limit;	// just counting arcs

    // Now start numbering the children
    // We start from the last arc to have more compression
    for (int j = limit - 1; j >= 0; --j) {
      pp = p->children[j].child;
      if (pp && pp->arc_no == -1) {
	pp->number_arcs(gtl);
        if (no_of_children && j == limit - 1 && nis_next_node(p->children[p->no_of_children - 1].child)) next_nodes++;
      }
    }
  }
  else arc_no = 0;
  return arc_no;
}

/* Purpose:	Writes arcs of a node and arcs of its descendants to a file.
 * Parameters:	outfile		- (o) output file.
 * Returns:	TRUE if arcs written, FALSE otherwise.
 * Remarks:	It is assumed that arcs were numbered with node::number_arcs
 *		with root being #1. #0 is a dummy arc. */
int node::write_arcs(ostream &outfile) {
  char		output_arc[10];
  char		*oa = &output_arc[0];
  fsa_arc_ptr	*dummy;
  int		i;
  node		*p;
  arc_node	pp;
  int		limit = no_of_children;

  // Establish the shape of the node and its arcs
  int gtl = dummy->gtl;
  int size = gtl + goto_offset;

  if (hit_count > 0) {
    hit_count = 0;		// mark the node as visited (written)
    if (no_of_children > 0) {
      for (i = 0; i < limit; i++) { // For all arcs of this node write the arcs
	pp = children[i]; 
	p = pp.child;
	*oa = pp.letter; // set the label
	// Set goto field
	if (p) {
	  int r = p->arc_no;
	  r <<= 3;
	  for (int i2 = 0; i2 < gtl;  i2++) {
	    oa[goto_offset + i2] = r & 0xff;
	    r >>= 8;
	  }
	}
	else for (int i3 = 0; i3 < gtl; i3++) oa[goto_offset + i3] = 0; // The address is NULL, so put all zeros

	fsa_set_next(oa, i == limit - 1 && wis_next_node(p) ? 1 : 0);
	if (wis_next_node(p) && i == limit - 1)
	  oa[goto_offset] |= 0x80; // make sure goto != 0

	if (p) {
	  // Set the number of arcs or the last arc marker
	  fsa_set_last(oa, i == limit - 1);
// This is exactly the opposite!!!!!
#if MAX_ARCS_PER_NODE==255
	  if (p->no_of_children > MAX_ARCS_PER_NODE) {
	    cerr << "Too many arcs leading from one node\nYour language "
	         << "contains more than " << MAX_ARCS_PER_NODE << " characters!\n";
	    exit(99);
	  }
#endif
	  fsa_set_final(oa, pp.is_final); // And the final bit
	}
	else {
	  // Set goto, number of arcs, and final fields
	  fsa_set_last(oa, i == limit - 1);
	  fsa_set_final(oa, TRUE);
	}

	// Write the result (the arc)
	int bytes_to_write = (p && i == limit - 1 && wis_next_node(p)) ? goto_offset + 1 : size;
	if (!(outfile.write(oa, bytes_to_write))) return FALSE;
      }//for i

      // For all arcs of this node or its brother
      // write arcs of descendants
      for (i = limit - 1; i >= 0; --i) {
	p = children[i].child;
	if (p) {
	  if (p->big_brother) p = p->big_brother;
	  // This must be done twice, as a 1-arc node may refer to 2-arc node
	  // merged with another 2-arc node
	  if (p->big_brother) p = p->big_brother;
	  if (!(p->write_arcs(outfile))) return FALSE;
	}//if p
      }//for i
    }//if no_of_children > 0
  }//if hit_count > 0
  return TRUE;
}

/* Purpose:	Prints node details for debugging purposes.
 * Parameters:	n		- (i) the node to be examined. */
void print_node(const node *n) {
  arc_node	*an;

  cerr << "Node: " << hex << (long)n << " hc=" << dec << ((node *)n)->hit_node(0)
       << " arc_no=" << n->get_arc_no() << " kids=" << n->get_no_of_kids();
  if (n->get_big_brother()) cerr << " bb=" << hex << (long)(n->get_big_brother()) << dec;
  cerr << endl;
  an = n->get_children();
  for (int i = 0; i < n->get_no_of_kids(); i++, an++)
    cerr << " +- " << an->letter << (an->is_final ? "!" : " ") << hex << (long)(an->child) << dec << endl;
}

/* Purpose:	Deletes a subgraph beginning at the node.
 * Parameters:	n	- (i/o) starting node of the subgraph to delete.
 * Remarks:	Only nodes with hit_count greater than zero are deleted.
 *		There is no need to deregister deleted nodes.
 *		Only unique nodes are registered.
 *		Only non-unique nodes are deleted. */
void delete_branch(node *branch) {
  arc_node	*an;
  if (branch->hit_node(0) == -1) return; // branch is to be deleted higher up
  if (branch->hit_node(0)) branch->hit_node(-1); // the branch is still in use
  else {
    an = branch->get_children();
    branch->hit_node(-1 - branch->hit_node(0));	// to avoid loops
    for (int i = 0; i < branch->get_no_of_kids(); i++, an++) if (an->child) delete_branch(an->child);
    delete branch;
  }
}

/* Purpose:	Mark nodes with a specified value of arc_no.
 * Parameters:	n	- (i) node to be marked;
 *		no	- (i) arc_no for that node.
 * Returns:	n - the node to be marked.
 * Remarks:	It is assumed that node with arc_no = no are already marked. */
node * mark_inner(node *n, const int no) {
  if (n->get_arc_no() == no) return n;

  n->set_arc_no(no);
  arc_node *an = n->get_children();
  for (int i = 0; i < n->get_no_of_kids(); i++, an++) if (an->child) mark_inner(an->child, no);
  return n;
}

/* Purpose:	Create a chain of nodes with each node holding a consecutive
 *		letter from postfix. Link it to the current node.
 * Parameters:	postfix		- (i) the string to be added.
 * Returns:	Pointer to this node. */
node * node::add_postfix(const char *postfix) {
  char	node_char;

  node_char = *postfix++;
  node *n = (*postfix ? new node() : (node *)NULL);
  add_child(node_char, (*postfix ? n->add_postfix(postfix) : (class node *)NULL));
  return this;
}

/* Purpose:	Add a child (an arc) to the given node.
 * Parameters:	letter		- (i) arc label;
 *		next_node	- (i) a node that the arc leads to or NULL for final node.
 * Returns:	Pointer to this node.
 * Remarks:	There may be an arc of this node labelled `letter'.
 *		In that case, use that arc (i.e., attach next node to it).
 *		If not, create a new arc. */
node * node::add_child(char letter, node *next_node) {
  arc_node	*new_children;
  arc_node	*new_child;
  int		i;

  // find if there is already an arc labelled `letter'
  // if so, put `next_node' at the end of that arc
  for (i = 0; i < no_of_children; i++) {
    if (children[i].letter == letter) {
      // arc found
      if (children[i].child) {
	cerr << "Two arcs from the same node with the same label!\n";
	cerr << "Trying to add an arc labeled `" << letter << "' leading to:" << endl;
	print_node(next_node);
	cerr << "to the following node:" << endl;
	print_node(this);
      }
      children[i].child = next_node;
      return this;
    }
  }

  // a new arc must be created
  new_children = new arc_node[no_of_children + 1];
  for (i = 0; i < no_of_children; i++) new_children[i] = children[i];

  new_child = new_children + no_of_children;
  new_child->child = next_node;
  new_child->is_final = (next_node == NULL);
  new_child->letter = letter;

  if (no_of_children) delete [] children;
  children = new_children;
  no_of_children++;

  return this;
}


static tree_index	primary_index = { {NULL}, 0 };
static tree_index	secondary_index = { {NULL}, 0 };

/* internal prototypes */
int match_tails(node *n, const int tail_size);

/* Purpose:	Delivers appropriate index.
 * Parameters:	index_name	- (i) index number.
 * Returns:	Root of required index.
 * Remarks:	Currently two indices available: PRIM_INDEX and SECOND_INDEX.
 *		No parameter checking. */
tree_index * get_index_by_name(const int index_name) { return (index_name ? &secondary_index : &primary_index); }

/* Purpose:	Looks for an appropriate entry at the given level.
 *		If not found, then creates it.
 * Parameters:	discriminator	- (i) value that should be matched against `data';
 *		table_node	- (i/o) a node in the table of the upper level that points to the table that is to be searched;
 *		to_register	- (i) TRUE means register, otherwise find and DO NOT REGISTER;
 *		cluster_size	- (i) how many entries are in the table at this level.
 * Returns:	A pointer to a table entry (existing previously or
 *		just created) pointing to the next (the lower) level.
 * Remarks:	It was necessary to add the cluster_size parameter in order
 *		to have the first level index on first arc label
 *		in secondary index. */
tree_index * register_at_level(const int discriminator, tree_index &table_node, const int to_register, const int cluster_size) {
  tree_index	*tip;
  int		i;

  if (table_node.counter == 0) {
    if (!to_register) return NULL;
    tip = table_node.down.indx = new tree_index[cluster_size + 1];
    table_node.counter = cluster_size + 1;
    for (i = 0; i <= cluster_size; i++) {
      tip[i].counter = 0;
      tip[i].allocated = 0;
      tip[i].down.indx = NULL;
    }
  }
  tip = table_node.down.indx;
  return ((tip[discriminator].counter || to_register) ? tip + discriminator : (tree_index *)NULL);
}

/* Class:	None (friend of class node).
 * Purpose:	Find if two subgraphs (represented by their root nodes)
 *		are isomorphic.
 * Parameters:	node1		- (i) first node;
 *		node2		- (i) second node;
 * Returns:	0		- if nodes are isomorphic;
 *		negative value	- if `node2' should precede `node1';
 *		positive value	- if `node2' should follow `node1'.
 * Description:	Only the children are compared for letters, being final,
 *		and addresses of the nodes they lead to.
 *		If they are equal, addresses of children are compared.
 * Remarks:	The number of children for the two nodes is not checked,
 *		as it assumed that the function is invoked only when
 *		the tree search has selected nodes of equal number of children
 *		for the comparison.
 *
 *		It is assumed that children are sorted alphabetically.
 *		This should be true if data for the automaton is sorted
 *		alphabetically.
 *
 *		It is possible to compare addresses of children, because
 *		children are registered first, and all child nodes
 *		of the node to be registered (if any) are already registered.
 * 
 *              One cannot say, which address is `bigger' than another,
 *              so the function returns `node1 < node2', because this
 *              triggers comparing next nodes at register_at_level. */
int cmp_nodes(const node *node1, const node *node2) {
  int		c;
  long int	l;
  arc_node	*p1, *p2;

  // compare children
  p1 = node1->children; p2 = node2->children;
  for (int i = 0; i < node1->no_of_children; i++, p1++, p2++)
    if ((l = ((long int)(p1->child) - (long int)(p2->child))) != 0) return ((l < 0) ? -1 : 1);
    else if ((c = p1->letter - p2->letter) != 0) return c;
    else if ((c = p1->is_final - p2->is_final) != 0) return c;

  // here nodes are isomorphic
  return 0;
}

/* Purpose:	Find either a node (or a subnode), or a place where it should
 *		be inserted in a register.
 * Parameters:	n		- node to be found;
 *		low		- address of the first pointer to node in register page;
 *		high		- address of the last pointer to node in register page;
 *		cmp		- function that compares two nodes;
 *		offset          - for pseudonodes, the number of the first arc of the pseudonode;
 *		group_size	- for pseudonodes, number of arcs in the pseudonode.
 * Returns:	Address of a pointer to node n, or a place where that pointer
 *		should be placed. */
node ** bsearch_reg(const node *n, node **low, node **high,
	    int (*cmp)(const node *, const node *, int, int),
	    int offset, int group_size) {
  node	**mid;
  int	c;

  while (low <= high) {
    mid = low + ((high - low) / 2);
    if ((c = cmp(n, *mid, offset, group_size)) < 0) high = mid - 1;
    else if (c > 0) low = mid + 1;
    else return mid;
  }
  return high + 1;
}

/* Purpose:	Finds an isomorphic subgraph in the automaton or registers a new node.
 * Parameters:	n		- (i) node to be found or registered;
 *		index_root	- (i/o) which index to use;
 *		register_it	- (i) if TRUE, register the node, if FALSE, find an isomorphic node.
 * Returns:	Pointer to an isomorphic node - if a node isomorphic to n found in the index, or NULL otherwise.
 * Remarks:	The index has levels indexed for:
 *		number of children (primary index) / first letter (secondary)
 *		hash function
 *		all features of the node.
 *
 *		It is assumed, that this function is not used for registering
 *		nodes in the secondary index. In theory it could, but the
 *		secondary index registers "subnodes" or "pseudonodes", not
 *		regular nodes, and no provision is taken here to take that
 *		fact into consideration. However, what is looked up in the
 *		secondary nodes are regular nodes, so the "this" pointer
 *		is always a regular node.
 *
 *		255 is the highest possible code of a character (so that
 *		it can be used in the first level index).
 *
 *		Changed:
 *		If register_it is true, if a node isomorphic to n is:
 *		found		- do not register, return isomorphic node;
 *		not found	- register, return NULL.
 *		This should speed up the code, as this function is
 *		rarely invoked for finding only, normally it is a
 *		sequence: find, and then register if not found. */
node * find_or_register(node *n, const int index_root_name, const int register_it) {
  tree_index	*tip;		// tree index pointer
  node		**current_node;	// subsequent entry in the leaf table
  int		not_found = TRUE;
  int		i;
  tree_index	*index_root;
  unsigned char	uc = (unsigned char)' ';

  index_root = get_index_by_name(index_root_name);

  // searching for the number of children entry
  if (n->get_no_of_kids()) uc = n->get_children()->letter;
  tip = (index_root_name
	 // secondary index
	 ? register_at_level(uc, *index_root, register_it, 255)
	 // primary index
	 : register_at_level(n->get_no_of_kids(), *index_root, register_it));
  if (tip == NULL) return NULL;

  // searching according to hash function value
  tip = register_at_level(n->hash(0, n->get_no_of_kids()), *tip, register_it);
  if (tip == NULL) return NULL;

  // now we search the lowest level, which has entries pointing directly
  // to end nodes
  not_found = TRUE;
  i = 0;
  current_node = tip->down.leaf;

  if (tip->down.leaf) {
    current_node = bsearch_reg(n, current_node, current_node + (tip->counter - 1),
			       (index_root_name ? part_cmp_nodes : a_cmp_nodes), -1, -1);
    not_found = !(current_node - tip->down.leaf < tip->counter &&
                  (index_root_name ? part_cmp_nodes(n, *current_node) : cmp_nodes(n, *current_node)) == 0);
  }
  i = current_node - tip->down.leaf;

  if (not_found) {
    if (!register_it) return NULL; // do not register

    // now `i' is the item number where we should put (`register') our new node
    current_node = tip->down.leaf;
    if (++(tip->counter) > tip->allocated) {
      node **p = new node *[tip->allocated += INDEX_SIZE_STEP];
      node **new_children = p;
      for (int j1 = 0; j1 < i; j1++) *p++ = *current_node++;
      *p++ = n;
      for (int j2 = i + 1; j2 < tip->counter; j2++) *p++ = *current_node++;
      if (tip->allocated > INDEX_SIZE_STEP) delete [] tip->down.leaf;
      tip->down.leaf = new_children;
    }
    else {
      for (int j = tip->counter - 1; j > i; --j) current_node[j] = current_node[j-1];
      current_node[i] = n;
    }

    // node registered
    return NULL;	// because node not found
  }//if not found

  // now `current_node->child' points to a node isomorphic to n
  if (register_it) return *current_node;
  return *current_node;
}

/* Class:	None (friend of class node).
 * Purpose:	Looks if it is possible to replace arcs of the smaller node
 *		with a subset of arcs of a larger node. If so, and replace is
 *		TRUE, it is done.
 * Parameters:	small_node	- (i/o) node to be replaced;
 *		big_node	- (i) node that may contain the same arcs;
 *		offset		- (i) (optional) if present, take to comparison only arcs #offset from small_node;
 *		group_size	- (i) (optional) how many arcs to consider.
 * Returns:	< 0 if group_size arcs of small_node from offset should precede the appropriate arcs of big_node in a list;
 *		= 0 if group_size arcs of small_node from offset are equal to the appropriate arcs of big_node;
 *		> 0 if group_size arcs of small_node from offset should follow the appropriate arcs of big_node in a list.
 * Returns[.h]: Number of the arc of big_node that is the first of
 *              small_node->no_of_children arcs that are identical to the arcs
 *              of small_node, or -1 if such arcs not found.
 * Remarks:	Let the small node have arcs (c d f) at offset.
 *		If the big node is (b c d f g) we compare:
 *		(b c d f g)  (b c d f g)  (a b c f g)  (a b c f g)
 *		(c d f)        (c d f)<-found! (return 0 [not mentioned in .h])
 *
 *		But for the big node being (b c d e f g):
 *		(b c d e f g)  (b c d e f g)  (b c d e f g)  (b c d e f g)
 *		(c d f)          (c d f)          (c d f)          (c d f)
 *		it does not work. [the rest of the paragraph not mentioned in .h]
 *		First, the nodes are aligned, then compared arc after arc:
 *		(b c d e f g)
 *		  (c d f)
 *		       \--difference found (returning > 0)
 *
 *		Maybe by chaging the order of arcs more arcs could be slashed. */
int part_cmp_nodes(const node *small_node, const node *big_node, const int offset, int group_size) {
  arc_node	*sp, *bp;
  int		big_offset;
  int		os = offset;
  int		gs = group_size;

  if (os == -1) {
    os = 0;
    gs = small_node->no_of_children;
  }
  int l = big_node->no_of_children - gs;
  bp = big_node->children;
  sp = small_node->children + os;

  // Find which arcs of big_node form a pseudonode
  for (big_offset = 0; big_offset <= l; big_offset++, bp++) if (sp->letter == bp->letter) break;

  // Check for errors (this can be removed later for speed
  if (sp->letter != bp->letter) {
    cerr << "Error in secondary index\n";
    cerr << "First arc label of the smaller node not found in the bigger one" << endl;
    cerr << "Smaller node:" << endl;
    print_node(small_node);
    cerr << "Bigger node:" << endl;
    print_node(big_node);
    exit (-20);
  }

  // Now compare arc by arc
  for (int i = 0; i < gs; i++, sp++, bp++) {
    if (sp->child != bp->child) return (sp->child < bp->child ? -1 : 1);
    else if (sp->letter != bp->letter) return (sp->letter - bp->letter);
    else if (sp->is_final != bp->is_final) return (sp->is_final - bp->is_final);
    //else compare more arcs...
  }

  // Here the subnodes must be equal
  return 0;
}
